unit RXMain;

{ This program provides an example of how to use the TreeView and ListView
  components in a fashion similar to the Microsoft Windows Explorer.

  It is not intended to be a fully functional resource viewer. }

{ WARNING / DISCLAIMER }

{ It has been extended in order to easily see/extract forms out of resources.
  But it is for educational purposes only. It is provided AS IS.
  Use AT YOUR OWN RISK if you want.
  Copyright 1996,1998 by Borland International
  Additional material   1999 by Franois GAILLARD }

{#ToDo2 - replace StringViewer by an extended memo to handle big forms }
{#ToDo1 - check (skipvalues) the unit generator to build also the *.pas (with a menu option) }

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls, Forms, Dialogs,
  ExeImage, StdCtrls, Buttons, ExtCtrls, ComCtrls, Menus, RXMisc, HexDump,
  ImgList;

type
  TMainForm = class(TForm)
    StatusBar: TStatusBar;
    TreeViewPanel: TPanel;
    Panel1: TPanel;
    ImageViewer: TImage;
    ListView: TListView;
    TreeView: TTreeView;
    Splitter: TPanel;
    Notebook: TNotebook;
    ListViewPanel: TPanel;
    ListViewCaption: TPanel;
    FileOpenDialog: TOpenDialog;
    FileSaveDialog: TSaveDialog;
    MainMenu: TMainMenu;
    miFile: TMenuItem;
    miFileExit: TMenuItem;
    miFileSave: TMenuItem;
    miFileOpen: TMenuItem;
    miView: TMenuItem;
    miViewStatusBar: TMenuItem;
    miViewLargeIcons: TMenuItem;
    miViewSmallIcons: TMenuItem;
    miViewList: TMenuItem;
    miViewDetails: TMenuItem;
    miHelp: TMenuItem;
    miHelpAbout: TMenuItem;
    Small: TImageList;
    Large: TImageList;
    StringViewer: TMemo;
    N1: TMenuItem;
    miFormsasText: TMenuItem;
    N2: TMenuItem;
    miWriteFormtoFile: TMenuItem;
    FormSaveDialog: TSaveDialog;
    procedure FileExit(Sender: TObject);
    procedure FileOpen(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ListViewEnter(Sender: TObject);
    procedure SaveResource(Sender: TObject);
    procedure SelectListViewType(Sender: TObject);
    procedure ShowAboutBox(Sender: TObject);
    procedure ToggleStatusBar(Sender: TObject);
    procedure TreeViewChange(Sender: TObject; Node: TTreeNode);
    procedure SplitterMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SplitterMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure SplitterMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ViewMenuDropDown(Sender: TObject);
    procedure NotebookEnter(Sender: TObject);
    procedure miFormsasTextClick(Sender: TObject);
    procedure miWriteFormtoFileClick(Sender: TObject);
  private
    FExeFile: TExeImage;
    HexDump: THexDump;
    SplitControl: TSplitControl;
    procedure LoadResources(ResList: TResourceList; Node: TTreeNode);
    procedure DisplayResources;
    procedure UpdateViewPanel;
    procedure UpdateListView(ResList: TResourceList);
    procedure UpdateStatusLine(ResItem: TResourceItem);
  end;

var
  MainForm: TMainForm;

implementation

uses About, RXTypes, consts, UConvert;

{$R *.DFM}
{$R RXIMAGES.RES}

resourcestring
  SNoResSelected = 'No resource selected';
  SNotAFormRes   = 'Not A Form Resource';
  SFormCaption   = 'Resource Explorer';
  SSaveFilter    = 'Other Resource (*.*)|*.*|Bitmaps (*.BMP)|*.BMP|'+
                   'Icons (*.ICO)|*.ICO|Cursor (*.CUR)|*.CUR';
//  SDefaultFilter = 'Tous les fichiers (*.*)|*.*';
  SFormSaveFilter= 'Delphi Forms (*.DFM)|*.DFM|'; {SDefaultFilter added later}
  SOpenFilter    = 'Executable File Images (*.EXE;*.DLL)|*.EXE;*.DLL|';  {SDefaultFilter added later}
  SFileExists    = 'File %s Exists.';
  SOverWrite     = 'Overwrite ?';

const
  itBitmap: TResType = ImgList.rtBitmap; // Reference for duplicate identifier
  ImageMap: array[TResourceType] of Byte = (2,4,5,3,2,2,2,2,2,2,2,2,2,2,2,2,2);
  ResFiltMap: array[TResourceType] of Byte = (1,4,2,3,1,1,1,1,1,1,1,1,1,1,1,1,1);


{ Utility Functions }

procedure Error(const ErrMsg: string);
begin
  raise Exception.Create(ErrMsg);
end;

procedure ErrorFmt(const ErrMsg: string; Params: array of const);
begin
  raise Exception.Create(format(ErrMsg, Params));
end;

function Confirm(const AMsg: String): Boolean;
begin
  Result := MessageDlg(AMsg, mtConfirmation, mbYesNoCancel, 0) = idYes;
end;

{ Non Event Handlers }

procedure TMainForm.LoadResources(ResList: TResourceList; Node: TTreeNode);
var
  I: Integer;
  CNode: TTreeNode;
begin
  for I := 0 to ResList.Count - 1 do
    with ResList[I] do
    begin
      CNode := TreeView.Items.AddChildObject(Node, Name, ResList[I]);
      if IsList then
      begin
        CNode.SelectedIndex := 1;
        LoadResources(List, CNode);
      end else
      begin
        CNode.ImageIndex := ImageMap[ResList[I].ResType];
        CNode.SelectedIndex := CNode.ImageIndex;
      end;

    end;
end;

procedure TMainForm.DisplayResources;
begin
  ListView.Items.Clear;
  TreeView.Selected := nil;
  TreeView.Items.Clear;
  LoadResources(FExeFile.Resources, nil);
  Caption := Format('%s - %s', [SFormCaption, AnsiLowerCaseFileName(FExeFile.FileName)]);
  with TreeView do
  begin
    SetFocus;
    Selected := Items[0];
  end;
end;

procedure TMainForm.UpdateViewPanel;
var
  R: TResourceItem;

  HeaderSize: Integer;
  Header: array[0..79] of Char;
  MemOutput: TMemoryStream;
  StrOutPut: TStringStream;
  s:string;
begin
  miWriteFormtoFile.Enabled := False;
  with TreeView do
  begin
    if Visible and Assigned(Selected) then
    begin
      R := TResourceItem(Selected.Data);
      if R.IsList then UpdateListView(R.List) else
      begin
        case R.ResType of
          rtBitmap, rtIconEntry, rtCursorEntry:
            begin
              ImageViewer.Picture.Assign(R);
              Notebook.PageIndex := 1;
            end;
          rtString, rtMenu:
            begin
              StringViewer.Lines.Assign(R);
              StringViewer.SelStart := 0;
              Notebook.PageIndex := 2;
            end
          else
            begin
              HexDump.Address := R.RawData;
              HexDump.DataSize := R.Size;
//------  FGD 23/11/99 13:34
              { test if Form <=> we have TPF0 as 1st positions in the RCDATA }
              if R.IsForm and (miFormsAsText.checked)then begin
                miWriteFormtoFile.Enabled := True;
                { build the form header }
                { prefix }
                Byte((@Header[0])^) := $FF;
                Word((@Header[1])^) := 10;
                { FormName }
                HeaderSize := StrLen(StrUpper(StrPLCopy(@Header[3], R.Name, 63))) + 10;
                Word((@Header[HeaderSize - 6])^) := $1030;
                { size }
                Longint((@Header[HeaderSize - 4])^) := R.Size;
                int64((@Header[HeaderSize - 2])^) := 0;
                { load into a meorystream }
                MemOutPut := TMemoryStream.Create;
                try
                  MemOutPut.WriteBuffer(Header, HeaderSize);
                  MemOutPut.Write(R.RawData^, R.Size);
                  MemOutPut.Position:=0;
                  { copy as text into a string stream }
                  StrOutPut:=TStringStream.Create(S);
                  try
                    ObjectResourceToText(MemOutPut,StrOutPut);
                    { show it into the StringViewer }
                    Notebook.PageIndex := 2;
                    StringViewer.Lines.Text:=StrOutPut.DataString;
                  finally
                    StrOutPut.Free;
                  end;
                finally
                  MemOutPut.free;
                end;
              end else begin
                Notebook.PageIndex := 3;
              end;
//-------
            end;
         end;
      end;
      UpdateStatusLine(R);
    end;
  end;
end;

procedure TMainForm.UpdateListView(ResList: TResourceList);
var
  I: Integer;
begin
  ListView.Items.Clear;
  for I := 0 to ResList.Count-1 do
    with ResList[I], ListView.Items.Add do
    begin
      Data := ResList[I];
      Caption := Name;
      SubItems.Add(Format('%.7x', [Offset]));
      SubItems.Add(Format('%.5x', [Size]));
      ImageIndex := ImageMap[ResType];
    end;
  Notebook.PageIndex := 0;
end;

procedure TMainForm.UpdateStatusLine(ResItem: TResourceItem);
begin
  if ResItem.IsList then
  begin
    ListViewCaption.Caption := ' '+TreeView.Selected.Text;
    StatusBar.Panels[0].Text := Format(' %d object(s)', [ListView.Items.Count]);
    StatusBar.Panels[1].Text := Format(' Offset: %x', [ResItem.Offset]);
  end else
  with ResItem do
  begin
    ListViewCaption.Caption := Format(' %s: %s', [ResTypeStr, Name]);
    StatusBar.Panels[0].Text := '';
    StatusBar.Panels[1].Text := Format(' Offset: %x  Size: %x', [Offset, Size]);
  end;
end;

{ Form Initialization }

procedure TMainForm.FormCreate(Sender: TObject);
begin
  SplitControl := TSplitControl.Create(Self);
  HexDump := CreateHexDump(TWinControl(NoteBook.Pages.Objects[3]));
  FileOpenDialog.Filter := SOpenFilter+SDefaultFilter;                  {FGD 23/11/99 9:49}
  FileSaveDialog.Filter := SSaveFilter;
  FormSaveDialog.Filter := SFormSaveFilter+SDefaultFilter;              {FGD 23/11/99 9:16}
  Small.ResourceLoad(itBitmap, 'SmallImages', clOlive);
  Large.ResourceLoad(itBitmap, 'LargeImages', clOlive);
  Notebook.PageIndex := 0;
  if (ParamCount > 0) and FileExists(ParamStr(1)) then
  begin
    Show;
    FExeFile := TExeImage.CreateImage(Self, ParamStr(1));
    DisplayResources;
  end;
end;

{ Menu Event Handlers }

procedure TMainForm.FileOpen(Sender: TObject);
var
  TmpExeFile: TExeImage;
begin
  with FileOpenDialog do
  begin
    if not Execute then Exit;
    TmpExeFile := TExeImage.CreateImage(Self, FileName);
    if Assigned(FExeFile) then FExeFile.Destroy;
    FExeFile := TmpExeFile;
    DisplayResources;
  end;
end;

procedure TMainForm.FileExit(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.ListViewEnter(Sender: TObject);
begin
  with ListView do
    if (Items.Count > 1) and (Selected = nil) then
    begin
      Selected := Items[0];
      ItemFocused := Selected;
    end;
end;

procedure TMainForm.SaveResource(Sender: TObject);
var
  ResItem: TResourceitem;

  function TreeViewResourceSelected: Boolean;
  begin
    Result := Assigned(TreeView.Selected) and
              Assigned(TreeView.Selected.Data) and
              not TResourceItem(TreeView.Selected.Data).IsList;
    if Result then ResItem := TResourceItem(TreeView.Selected.Data);
  end;

  function ListViewResourceSelected: Boolean;
  begin
    Result := Assigned(ListView.Selected) and
              Assigned(ListView.Selected.Data) and
              not TResourceItem(ListView.Selected.Data).IsList;
    if Result then ResItem := TResourceItem(ListView.Selected.Data);
  end;

begin
  if TreeViewResourceSelected or ListViewResourceSelected then
  with FileSaveDialog do
  begin
    FilterIndex := ResFiltMap[ResItem.ResType];
    if Execute then
      ResItem.SaveToFile(FileName)
  end
  else
    Error(SNoResSelected);
end;

procedure TMainForm.SelectListViewType(Sender: TObject);
begin
  ListView.ViewStyle := TViewStyle(TComponent(Sender).Tag);
end;

procedure TMainForm.ShowAboutBox(Sender: TObject);
begin
  About.ShowAboutBox;
end;

procedure TMainForm.ToggleStatusBar(Sender: TObject);
begin
  StatusBar.Visible := not StatusBar.Visible;
end;

procedure TMainForm.TreeViewChange(Sender: TObject; Node: TTreeNode);
begin
  UpdateViewPanel;
end;

procedure TMainForm.ViewMenuDropDown(Sender: TObject);
var
  I: Integer;
begin
  miViewStatusBar.Checked := StatusBar.Visible;
  for I := 0 to miView.Count-1 do
    with miView.Items[I] do
      if GroupIndex = 1 then
        Checked := (Tag = Ord(ListView.ViewStyle));
end;

procedure TMainForm.SplitterMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if (Button = mbLeft) and (Shift = [ssLeft]) then
    SplitControl.BeginSizing(Splitter, TreeViewPanel);
end;

procedure TMainForm.SplitterMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
  with SplitControl do if Sizing then ChangeSizing(X, Y);
end;

procedure TMainForm.SplitterMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  with SplitControl do if Sizing then EndSizing;
end;

procedure TMainForm.NotebookEnter(Sender: TObject);
var
  Page: TWinControl;
begin
  with NoteBook do
  begin
    Page := TWinControl(Pages.Objects[PageIndex]);
    if (Page.ControlCount > 0) and (Page.Controls[0] is TWinControl) then
      TWinControl(Page.Controls[0]).SetFocus;
  end;
end;

procedure TMainForm.miFormsasTextClick(Sender: TObject);
begin
  with Sender as TMenuItem do
    Checked:= not Checked;
end;


procedure TMainForm.miWriteFormtoFileClick(Sender: TObject);
{ FGD 23/11/99 13:33 }
var
  R: TResourceItem;
  HeaderSize: Integer;
  Header: array[0..79] of Char;
  Output,OutPutPas: TFileStream;
begin
  with TreeView do begin
    if Visible and Assigned(Selected) then begin
      R := TResourceItem(Selected.Data);
      if R.IsList then
        Error(SNotAFormRes)
      else begin
        case R.ResType of
          rtBitmap, rtIconEntry, rtCursorEntry, rtString, rtMenu:
            Error(SNotAFormRes)
          else begin
            { test if Form <=> we have TPF0 as 1st positions in the RCDATA }
            if R.IsForm then begin
              { build the form header }
              { prefix }
              Byte((@Header[0])^) := $FF;
              Word((@Header[1])^) := 10;
              { FormName }
              HeaderSize := StrLen(StrUpper(StrPLCopy(@Header[3], R.Name, 63))) + 10;
              Word((@Header[HeaderSize - 6])^) := $1030;
              { size }
              Longint((@Header[HeaderSize - 4])^) := R.Size;
              int64((@Header[HeaderSize - 2])^) := 0;
              { we need a file }
              with FormSaveDialog do begin
                FileName:=R.Name + '.DFM';
                PChar(FileName)[0]:='U';
                if not Execute then
                  Exit;
                OutPut := TFileStream.Create(FileName,fmCreate);
              end;
              { write to file }
              try
                OutPut.WriteBuffer(Header, HeaderSize);
                OutPut.Write(R.RawData^, R.Size);
                OutPut.Position:=0;
                OutPutPas:= TFileStream.Create(ChangeFileExt(FormSaveDialog.FileName,'.PAS'),fmCreate);
                try
                  { conversion to Pascal. utility from UConvert }
                  ObjectResourceToPascal(OutPut,OutPutPas,nil,ChangeFileExt(FormSaveDialog.FileName,'.PAS'));
                finally
                  OutPutPas.Free;
                end;
              finally
                OutPut.free;
              end;
            end else begin
              Error(SNotAFormRes);
            end; { if }
          end; { case else }
        end; { case }
      end; { if isList}
    end; { if treeview ... selected }
  end; { with treeview }
end;

end.
