unit UConvert;

interface

uses Classes;

procedure ObjectBinaryToText(Input, Output: TStream; ComponentsList:TStrings; fname: string);
procedure ObjectResourceToPascal(Input, Output: TStream; ComponentsList:TStrings; fname: string);

implementation

uses Consts, SysUtils;

{ Resource to text conversion }

procedure ObjectResourceToPascal(Input, Output: TStream; ComponentsList:TStrings; fname: string);
begin
  Input.ReadResHeader;
  ObjectBinaryToText(Input, Output, ComponentsList, fname);
end;

procedure SkipValue(AReader:TReader);

  procedure SkipList;
  begin
    while not AReader.EndOfList do SkipValue(AReader);
    AReader.ReadListEnd;
  end;

  procedure SkipBytes(Count: Longint);
  var
    Bytes: array[0..255] of Char;
  begin
    while Count > 0 do
      if Count > SizeOf(Bytes) then
      begin
        AReader.Read(Bytes, SizeOf(Bytes));
        Dec(Count, SizeOf(Bytes));
      end
      else
      begin
        AReader.Read(Bytes, Count);
        Count := 0;
      end;
  end;

  procedure SkipBinary;
  var
    Count: Longint;
  begin
    AReader.Read(Count, SizeOf(Count));
    SkipBytes(Count);
  end;

  procedure SkipProperty;
  begin
    AReader.ReadStr; { Skips property name }
    SkipValue(Areader);
  end;

  procedure SkipCollection;
  begin
    while not AReader.EndOfList do
    begin
      if AReader.NextValue in [vaInt8, vaInt16, vaInt32] then SkipValue(AReader);
      SkipBytes(1);
      while not AReader.EndOfList do SkipProperty;
      AReader.ReadListEnd;
    end;
    AReader.ReadListEnd;
  end;

  procedure SkipSetBody;
  begin
    while AReader.ReadStr <> '' do begin end;
  end;

begin
  case AReader.ReadValue of
    vaNull: begin end;
    vaList: SkipList;
    vaInt8: SkipBytes(1);
    vaInt16: SkipBytes(2);
    vaInt32: SkipBytes(4);
    vaExtended: SkipBytes(SizeOf(Extended));
    vaString, vaIdent: AReader.ReadStr;
    vaFalse, vaTrue: begin end;
    vaBinary: SkipBinary;
    vaSet: SkipSetBody;
    vaCollection: SkipCollection;
  end;
end; {skipValue}


procedure CheckValue(AReader:TReader; Value: TValueType);
begin
  if AReader.NextValue <> Value then
  begin
    SkipValue(AReader);
    raise EReadError.Create(SInvalidPropertyValue);
  end
  else
    Areader.ReadValue;
end; {CheckValue}


{ Binary to text conversion }

procedure ObjectBinaryToText(Input, Output: TStream; ComponentsList:TStrings; fname:string);
var
  NestingLevel: Integer;
  SaveSeparator: Char;
  Reader: TReader;
  Writer: TWriter;
  Pos: integer;
  s1,s2: string;

  procedure WriteIndent;
  const
    Blanks: array[0..1] of Char = '  ';
  var
    I: Integer;
  begin
    for I := 1 to NestingLevel do begin
      Writer.Write(Blanks, SizeOf(Blanks));
    end;
  end;

  procedure WriteStr(const S: string);
  begin
    Writer.Write(S[1], Length(S));
  end;

  procedure NewLine;
  begin
    WriteStr(#13#10);
    WriteIndent;
  end;

  procedure ConvertValue; forward; //FGD

  procedure ConvertHeader( var ClassName, ObjectName: string; bFirst : boolean); //FGD
  var
    Flags: TFilerFlags;
    Position: Integer;
  begin
    Reader.ReadPrefix(Flags, Position);
    ClassName := Reader.ReadStr;
    ObjectName := Reader.ReadStr;
(*    bChanged := bToFile and (ComponentsList.IndexOf(ClassName)>=0); { true if in list}  //FGD
    if bChanged then                                     //FGD
      bToFile:=false;                                    //FGD*)
    if bFirst then begin
      NewLine;
      WriteStr( ClassName + ' = class( TForm) ');
      if ffInherited in Flags then
        WriteStr(' { Warning : inherited from descendant }');
      Inc(NestingLevel,1);
    end else
    if ObjectName <> '' then
    begin
      NewLine;
      writeStr( ObjectName + ' : ' + ClassName +';');
    end;
    if ffChildPos in Flags then
    begin
      WriteStr(' [');
      WriteStr(IntToStr(Position));
      WriteStr(']');
    end;
//    WriteStr(#13#10);
  end;                                                   //FGD

  procedure ConvertBinary;       //FGD
  const
    BytesPerLine = 32;
  var
    I: Integer;
    Count: Longint;
    Buffer: array[0..BytesPerLine - 1] of Char;
  begin
    Reader.ReadValue;
    Reader.Read(Count, SizeOf(Count));
    while Count > 0 do
    begin
      if Count >= 32 then I := 32 else I := Count;
      Reader.Read(Buffer, I);
      Dec(Count, I);
    end;
  end;

  procedure ConvertProperty(const ObjectName:string); forward;

  procedure ConvertValue;         //FGD
  const
    LineLength = 64;
  var
    S: string;
  begin
    case Reader.NextValue of
      vaList:
        begin
          Reader.ReadValue;
          while not Reader.EndOfList do
          begin
            ConvertValue;                       //FGD
          end;
          Reader.ReadListEnd;
        end;
      vaInt8, vaInt16, vaInt32:
        Reader.ReadInteger;
      vaExtended:
        Reader.ReadFloat;
      vaSingle:
        Reader.ReadSingle;
      vaCurrency:
        Reader.ReadCurrency;
      vaDate:
        Reader.ReadDate;
      vaWString:
        begin
          Reader.ReadWideString;
        end;
      vaString, vaLString:
        begin
          Reader.ReadString;
        end;
      vaIdent, vaFalse, vaTrue, vaNil, vaNull:
        Reader.ReadIdent;//FGD
      vaBinary:
        ConvertBinary;                          //FGD
      vaSet:
        begin
          Reader.ReadValue;
          while True do
          begin
            S := Reader.ReadStr;
            if S = '' then Break;
          end;
        end;
      vaCollection:
        begin
          Reader.ReadValue;
          while not Reader.EndOfList do
          begin
            if Reader.NextValue in [vaInt8, vaInt16, vaInt32] then
            begin
              ConvertValue;                     //FGD
            end;
            CheckValue(Reader,vaList);                   //FGD
            while not Reader.EndOfList do ConvertProperty('');//FGD
            Reader.ReadListEnd;
          end;
          Reader.ReadListEnd;
        end;
    end;
  end;

  procedure ConvertProperty(const ObjectName:string);
  begin
    Reader.ReadStr;
    ConvertValue;
  end;

  procedure ConvertObject( bFirst : boolean);         //FGD
  var
(*    PropertiesList : TStrings;                           //FGD*)
    ClassName,ObjectName: string;                                   //FGD
  begin
    ClassName := '';                                     //FGD
    ObjectName := '';                                     //FGD
    ConvertHeader( ClassName, ObjectName, bFirst);          //FGD
(*    PropertiesList := TStringList.Create;              //FGD
    if NestingLevel=0 then {TForm whatever is classname }
      IniF.ReadSection('TForm',PropertiesList)         //FGD
    else
      IniF.ReadSection(ClassName,PropertiesList);      //FGD
    try                                                  //FGD *)
      while not Reader.EndOfList do ConvertProperty(ObjectName);//FGD
      Reader.ReadListEnd;
      while not Reader.EndOfList do ConvertObject(False); //FGD
      Reader.ReadListEnd;
(*    finally                                              //FGD
      if bCreated then  { beware of recursion }
        PropertiesList.Free;                             //FGD
    end;                                                 //FGD
    if bChanged then                                     //FGD
      bToFile:=true;                                     //FGD*)
  end;

begin
  NestingLevel := 0;
  Reader := TReader.Create(Input, 4096);
  SaveSeparator := DecimalSeparator;
  DecimalSeparator := '.';
  try
    Writer := TWriter.Create(Output, 4096);
    try
      Reader.ReadSignature;
      WriteStr('unit '+ ChangeFileExt(ExtractFileName(fname),'') + ';'#13#10#13#10);
      WriteStr('{ Automatically generated }'#13#10#13#10);
      WriteStr('interface'#13#10#13#10);
      WriteStr('uses'#13#10'  SysUtils, Windows, Messages, Classes, Graphics, Controls, Forms, Dialogs;'#13#10#13#10);
      WriteStr('type');
      Inc(NestingLevel,1);
      Pos:=Reader.Position;
      ConvertObject(True);                          //FGD
      dec(NestingLevel,1);
      NewLine;
      WriteStr('end;');
      dec(NestingLevel,1);
      NewLine;
      NewLine;
      Reader.Position:= Pos;
      WriteStr('var ');
      Inc(NestingLevel,1);
      s1:='';
      s2:='';
      ConvertHeader(s1,s2,False);
      NewLine;
      WriteStr(#13#10'implementation'#13#10#13#10);
      WriteStr('{$R *.DFM}'#13#10#13#10);
      WriteStr('end.'#13#10#13#10);

    finally
      Writer.Free;
    end;
  finally
    DecimalSeparator := SaveSeparator;
    Reader.Free;
  end;
end;  {ObjectBinaryToText}


end.
